/* ------------------------------------------------------------- */
/* U78.cc                                                        */
/* Ayres Freitas (afreitas@pitt.edu)                             */
/* last revision 20.08.19                                        */
/* ------------------------------------------------------------- */
/* 3-loop self-energy master integrals w/ two sub-loop triangles */
/* ------------------------------------------------------------- */
/* Reference:                                                    */
/*   [2] = Bauberger, Freitas, Wiegand, arXiv:1908.09887         */
/* ------------------------------------------------------------- */

#include "U456.h"
#include "li.h"
#include "qagutil.h"
#include "B0.h"
#include "C0.h"
#include "clooptools.h"


static Double ps,p,m1s,m2s,m3s,m4s,m5s,m6s,m7s,m8s,ysav,qsav;
static Double prec,abserr;
static Cplx resfacx,resfacy;


/************************* U8a: ***************************/

// integrand of the inner (x) integration
static Cplx U8fct0(Double x)
{ 
  Double l = la(x,ysav,ps);
  if(l > Double(0))
  {
    ComplexType c0a[3], c0b[3];
    C0nocache(c0a,ps,x,ysav,m1s,m2s,m3s);
    C0nocache(c0b,ps,x,ysav,m4s,m5s,m6s);
    return(sqrt(l)/(2*ps)*Pi*Cplx(c0a[0]*c0b[0]));
  }
  else
    return(Cplx(0));
}

// treat pole at x=m7s through Cauchy PV, and add residue in integrand through 
// a smooth function that integrate to 1;
// this helps to control the integration precision and avoid large cancellations
// between integral and residue
static Cplx U8fct2(Double xp)
{ 
  return((U8fct0(m7s+xp) - U8fct0(m7s-xp))/xp - resfacx*exp(-sqr(xp)/(4*m7s))); 
}

// integrand of the outer (y) integration
static Cplx U8fct0pint(Double y)
{
  Double err1=0;
  ysav=y;
  resfacx = PiI*U8fct0(m7s) / sqrt(Pi*m7s);  // to account for residue at pole
  Cplx z = Cqaginf2(U8fct2,0,prec,&err1);
  clearcache();
  return(z);
}

// treat pole at y=m8s through Cauchy PV, and add residue in integrand through 
// a smooth function that integrate to 1;
// this helps to control the integration precision and avoid large cancellations
// between integral and residue
static Cplx U8fct2int(Double yp)
{ 
  return((U8fct0pint(m8s+yp) - U8fct0pint(m8s-yp))/yp - 
         resfacy*exp(-sqr(yp)/(4*m8s))); 
}

Cplx U8aint(Double p2, Double m1se, Double m2se, Double m3se, Double m4se, Double m5se, Double m6se, Double m7se, Double m8se,
            Double prece, Double *abserre)
{
  prec = prece;
  m1s=m2se; m2s=m1se; m3s=m3se; ps = p2;
  m4s=m8se; m5s=m7se; m6s=m6se; m7s = m4se; m8s = m5se;

  resfacy = PiI*U8fct0pint(m8s) / sqrt(Pi*m8s);  // to account for residue at pole
  Cplx z = Cqaginf(U8fct2int,0,100*prec,abserre);
  *abserre /= sqr(Pi);
  return(z/(I*Pi*Pi));
}


/************************* U7a: ***************************/

// integrand of the inner (x) integration
static Cplx U7fct0(Double x)
{ 
  if(x > sqr(qsav))
    return(Cplx(0));
  Double y = x + ps + 2*qsav*p;
  ComplexType c0a[3], c0b[3];
  C0nocache(c0a,ps,x,y,m1s,m2s,m3s);
  C0nocache(c0b,ps,x,y,m4s,m5s,m6s);
  Cplx z = (2*Pi*sqrt(sqr(qsav)-x)*Cplx(c0a[0]*c0b[0])); 
  return(z);
}

// treat pole at x=m7s through Cauchy PV, and add residue in integrand through 
// a smooth function that integrate to 1;
// this helps to control the integration precision and avoid large cancellations
// between integral and residue
static Cplx U7fct2(Double xp)
{ 
  return((U7fct0(m7s+xp) - U7fct0(m7s-xp))/xp - resfacx*exp(-sqr(xp)/(4*m7s))); 
}

// integrand of the outer (q0) integration
static Cplx U7fct0int(Double q0)
{
  Double err1=0;
  qsav=q0;
  resfacx = PiI*U7fct0(m7s) / sqrt(Pi*m7s);  // to account for residue at pole
  Cplx z = Cqaginf2(U7fct2,0,prec,&err1);
//  clearcache();
  return(z);
}

// map onto [0,infty) interval
static Cplx U7fct2int(Double q0)
{
  return(U7fct0int(q0) + U7fct0int(-q0));
}

Cplx U7aint(Double p2, Double m1se, Double m2se, Double m3se, Double m4se, 
            Double m6se, Double m7se, Double m8se, Double prece, Double *abserre)
{
  prec = prece;
  m1s=m2se; m2s=m1se; m3s=m3se; ps = p2; p = sqrt(p2);
  m4s=m8se; m5s=m7se; m6s=m6se; m7s = m4se;

  Cplx z = Cqaginf(U7fct2int,0,100*prec,abserre);
  *abserre /= sqr(Pi);
  return(z/(I*Pi*Pi));
}


/************************* U7a1: ***************************/

// integrand of the inner (x) integration:

Cplx DM3C0exp1(Double p10, Double p20, Double p12, Double m1, Double m2, Double m3)
{
  return(
    (2+ B0A(0,m3,m3) + B0A(p10,m1,m2) 
      - (1+(m1-m2)/p10)*B0A(p12,m1,m3) 
      - (1-(m1-m2)/p10)*B0A(p12,m2,m3))/sqr(p12));
}

static Cplx U71fct0(Double x)
{ 
  Cplx z;
  if(x > sqr(qsav))
    return(Cplx(0));
  Double y = x + ps + 2*qsav*p;
  ComplexType c0a[3];
  C0nocache(c0a,ps,x,y,m4s,m5s,m6s);
  if(abs(x/(qsav*p + ps)) > 1e6)
    z = (2*Pi*sqrt(sqr(qsav)-x)*
  	     DM3C0exp1(ps,x,y,m1s,m2s,m3s)*Cplx(c0a[0])); 
  else
    z = (2*Pi*sqrt(sqr(qsav)-x)*
  	     DM3C0(ps,x,y,m1s,m2s,m3s)*Cplx(c0a[0])); 
  return(z);
}

// treat pole at x=m7s through Cauchy PV, and add residue in integrand through 
// a smooth function that integrate to 1;
// this helps to control the integration precision and avoid large cancellations
// between integral and residue
static Cplx U71fct2(Double xp)
{ 
  return((U71fct0(m7s+xp) - U71fct0(m7s-xp))/xp - resfacx*exp(-sqr(xp)/(4*m7s))); 
}

// integrand of the outer (q0) integration
static Cplx U71fct0int(Double q0)
{
  Double err1=0;
  qsav=q0;
  resfacx = PiI*U71fct0(m7s) / sqrt(Pi*m7s);  // to account for residue at pole
  Cplx z = Cqaginf2(U71fct2,0,prec,&err1);
  clearcache();
  return(z);
}

// map onto [0,infty) interval
static Cplx U71fct2int(Double q0)
{
  return(U71fct0int(q0) + U71fct0int(-q0));
}

Cplx U7a1int(Double p2, Double m1se, Double m2se, Double m3se, Double m4se, 
             Double m6se, Double m7se, Double m8se, Double prece, Double *abserre)
{
  prec = prece;
  m1s=m2se; m2s=m1se; m3s=m3se; ps = p2; p = sqrt(p2);
  m4s=m8se; m5s=m7se; m6s=m6se; m7s = m4se;

  Cplx z = Cqaginf(U71fct2int,0,100*prec,abserre);
  *abserre /= sqr(Pi);
  return(z/(I*Pi*Pi));
}


/************************* U7a2: ***************************/

// integrand of the inner (x) integration:

Cplx DM3C0exp2(Double p10, Double p20, Double p12, Double m1, Double m2, Double m3)
{
  return(
  (2*m2*p20*B0A(0,m2,m2) + p20*(-m2 - m3 + p20)*B0A(0,m3,m3) - 
   p20*(m2 - m3 + p20)*(B0A(p20,m2,m3)-2))/
   (p12*p20*(sqr(m2) + sqr(m3 - p20) - 2*m2*(m3 + p20))));
}

static Cplx U72fct0(Double x)
{ 
  Cplx z;
  if(x > sqr(qsav))
    return(Cplx(0));
  Double y = x + ps + 2*qsav*p;
  ComplexType c0a[3];
  C0nocache(c0a,ps,x,y,m4s,m5s,m6s);
  if(abs(x/(qsav*p + ps)) > 1e6)
    z = (2*Pi*sqrt(sqr(qsav)-x)*
  	       DM3C0exp2(x,ps,y,m3s,m2s,m1s)*Cplx(c0a[0])); 
  else
    z = (2*Pi*sqrt(sqr(qsav)-x)*
  	       DM3C0(x,ps,y,m3s,m2s,m1s)*Cplx(c0a[0])); 
  return(z);
}


// treat pole at x=m7s through Cauchy PV, and add residue in integrand through 
// a smooth function that integrate to 1;
// this helps to control the integration precision and avoid large cancellations
// between integral and residue
static Cplx U72fct2(Double xp)
{ 
  return((U72fct0(m7s+xp) - U72fct0(m7s-xp))/xp - resfacx*exp(-sqr(xp)/(4*m7s))); 
}

// integrand of the outer (q0) integration
static Cplx U72fct0int(Double q0)
{
  Double err1=0;
  qsav=q0;
  resfacx = PiI*U72fct0(m7s) / sqrt(Pi*m7s);  // to account for residue at pole
  Cplx z = Cqaginf2(U72fct2,0,prec,&err1);
  clearcache();
  return(z);
}

// map onto [0,infty) interval
static Cplx U72fct2int(Double q0)
{
  return(U72fct0int(q0) + U72fct0int(-q0));
}

Cplx U7a2int(Double p2, Double m1se, Double m2se, Double m3se, Double m4se, 
             Double m6se, Double m7se, Double m8se, Double prece, Double *abserre)
{
  prec = prece;
  m1s=m2se; m2s=m1se; m3s=m3se; ps = p2; p = sqrt(p2);
  m4s=m8se; m5s=m7se; m6s=m6se; m7s = m4se;
  Double err1=0, err2=0;

  Cplx z = Cqaginf(U72fct2int,0,100*prec,abserre);
  *abserre /= sqr(Pi);
  return(z/(I*Pi*Pi));
}
