/* ------------------------------------------------------------- */
/* Umn.cc                                                        */
/* Ayres Freitas (afreitas@pitt.edu)                             */
/* last revision 06.08.19                                        */
/* ------------------------------------------------------------- */
/* 3-loop self-energy master integrals with one sub-loop bubble  */
/* and one sub-loop triangle (subtracted finite parts)           */
/* ------------------------------------------------------------- */
/* Reference:                                                    */
/*   [2] = Bauberger, Freitas, Wiegand, arXiv:1908.09887         */
/* ------------------------------------------------------------- */

#include "Umn.h"
#include "li.h"
#include "qagutil.h"
#include "B0.h"
#include "T345.h"
#include "tinline.h"

#define Errstream cerr

/******************************************************************/
/* Throughout this file the numbers are defined as Double or Cplx */
/* to enable a generalization to arbitrary precision              */
/******************************************************************/

/******************************************************************/
/* convention: vars is frequently used for (var)^2                */
/* for example: m1s denotes m_1^2.                                */
/******************************************************************/


/****** Self-energy integrals with single-power propagators: ******/

static Double ps,m1s,m2s,m3s,m4s,m5s,m6s,m7s;
static Double ASYMP_LIMIT_T5a;
static Cplx T5ex;
static Double prec,abserr;
static Cplx U6mk(Double x);
static Cplx U6nk(Double x);
static Cplx U7mk(Double x);


static Cplx U6mk(Double s1)
{
  Double err1=0;
  Cplx z;
  if(s1 > ASYMP_LIMIT_T5a)
  {
/* Use extrapolation of integrand for large s1, based on form of large-mass expansion
   (A + B*log(s1) + C*log^2(s1))/s1^2 + O(s1^-3)
   C = -ps/4
   B = -(m1s/2-m2s/2+m3s+m4s)*(B0fin(ps,m1s,m2s)-B0fin(0,m1s,m2s))
       -ps/2*(B0fin(ps,m1s,m2s)-1)
   A is determined approximatedly from value of integrand for s1=ASYMP_LIMIT_T5a
*/
    z = T5ex*sqr(ASYMP_LIMIT_T5a/s1) + log(ASYMP_LIMIT_T5a/s1)*
    	((.5*m1s-.5*m2s+m3s+m4s)*(B0A(ps,m1s,m2s)-B0A(0,m1s,m2s)) +
	 ps/2*(B0A(ps,m1s,m2s)+1) + ps/4*(log(s1) + log(ASYMP_LIMIT_T5a)))/
	sqr(s1);
    z *= -DeltaB0(s1,m6s,m7s);
  }
  else
    z = -DeltaB0(s1,m6s,m7s)*(T5aint(ps,m1s,m2s,m3s,m4s,s1,prec,&err1)
    			      - T5aint(0,m1s,m2s,m3s,m4s,s1,prec,&err1) + 1/s1*
		(T4aint(ps,m1s,m2s,m3s,m4s,prec,&err1) +
		 (B0A(ps,m1s,m2s)-B0A(0,m1s,m2s))*
		 (B0A(m1s,m3s,m4s)-B0A(0,0,s1))));
  return(z);
}

Cplx U6mint(Double pse, Double m1se, Double m2se, Double m3se, Double m4se,
            Double m6se, Double m8se, Double prece, Double *abserre)
{
  ps=pse; m1s=m1se; m2s=m2se; m3s=m3se; m4s=m4se; m6s=m6se; m7s=m8se;
  prec=prece*10;
  Double err;
  ASYMP_LIMIT_T5a = (abs(ps)+m1s+m2s+m3s+m4s)*1e8;
  T5ex = -U6mk(ASYMP_LIMIT_T5a)/DeltaB0(ASYMP_LIMIT_T5a,m6s,m7s);
  Cplx z=Cqaginf2(U6mk,sqr(sqrt(m6s)+sqrt(m7s)),prece*100,&err);
  *abserre = err;
  return(z);
}


static Cplx U6nk(Double s1)
{
  Double err1=0;
  Cplx z;
  if(s1 > ASYMP_LIMIT_T5a)
  {
/* Use extrapolation of integrand for large s1, based on form of large-mass expansion
   (A + B*log(s1) + C*log^2(s1))/s1^2 + O(s1^-3)
   C = ps/2
   B = -(m1s/2+m2s/2+m4s+m5s)*(B0fin(ps,m1s,m2s)-B0fin(0,m1s,m2s))
       -(m1s+m2s+m4s/2+m5s/2)*(B0fin(ps,m4s,m5s)-B0fin(0,m4s,m5s))
       +ps/2*(B0fin(ps,m1s,m2s)+B0fin(ps,m4s,m5s)-3)
   A is determined approximatedly from value of integrand for s1=ASYMP_LIMIT_T5a
*/
    z = T5ex*sqr(ASYMP_LIMIT_T5a/s1) + log(ASYMP_LIMIT_T5a/s1)*
    	((.5*m1s+.5*m2s+m4s+m5s)*(B0A(ps,m1s,m2s)-B0A(0,m1s,m2s)) +
	 (.5*m4s+.5*m5s+m1s+m2s)*(B0A(ps,m4s,m5s)-B0A(0,m4s,m5s)) 
	 - ps/2*(B0A(ps,m1s,m2s)+B0A(ps,m4s,m5s)+1) 
	 - ps/2*(log(s1) + log(ASYMP_LIMIT_T5a)))/sqr(s1);
    z *= -DeltaB0(s1,m6s,m7s);
  }
  else
    z = -DeltaB0(s1,m6s,m7s)*(T5aint(ps,m1s,m2s,s1,m4s,m5s,prec,&err1)
    			      - T5aint(0,m1s,m2s,s1,m4s,m5s,prec,&err1) + 1/s1*
	     (B0A(ps,m1s,m2s)*B0A(ps,m4s,m5s) - B0A(0,m1s,m2s)*B0A(0,m4s,m5s) -
	      B0A(0,0,s1)*(B0A(ps,m1s,m2s) + B0A(ps,m4s,m5s)
	      		   - B0A(0,m1s,m2s) - B0A(0,m4s,m5s))));
  return(z);
}

Cplx U6nint(Double pse, Double m1se, Double m2se, Double m3se, Double m6se,
            Double m7se, Double m8se, Double prece, Double *abserre)
{
  ps=pse; m1s=m1se; m2s=m2se; m4s=m7se; m5s=m8se; m6s=m3se; m7s=m6se;
  prec=prece*10;
  Double err;
  ASYMP_LIMIT_T5a = (abs(ps)+m1s+m2s+m4s+m5s)*1e8;
  T5ex = -U6nk(ASYMP_LIMIT_T5a)/DeltaB0(ASYMP_LIMIT_T5a,m6s,m7s);
  Cplx z=Cqaginf2(U6nk,sqr(sqrt(m6s)+sqrt(m7s)),prece*100,&err);
  *abserre = err;
  return(z);
}


static Cplx U7mk(Double s1)
{
  Double err1=0;
  Cplx z;
  if(s1 > ASYMP_LIMIT_T5a)
  {
/* Use extrapolation of T5 for large s1, based on form of large-mass expansion
   T5 = (A + B*log(s1) + C*log^2(s1))/s1 + O(s1^-2)
   C = -1/2
   B = -B0fin(ps,m1s,m2s)
   A is determined approximatedly from value of T5 for s1=ASYMP_LIMIT_T5a
*/
    z = T5ex*(ASYMP_LIMIT_T5a/s1) + log(ASYMP_LIMIT_T5a/s1)*
    	(B0A(ps,m1s,m2s)+2 + .5*log(s1) + .5*log(ASYMP_LIMIT_T5a))/s1;
    z *= -DeltaB0(s1,m6s,m7s);
  }
  else
    z = -DeltaB0(s1,m6s,m7s)*T5aint(ps,m1s,m2s,m3s,m4s,s1,prec,&err1);
  return(z);
}

Cplx U7mint(Double pse, Double m1se, Double m2se, Double m3se, Double m4se,
            Double m5se, Double m6se, Double m8se, Double prece, Double *abserre)
{
  ps=pse; m1s=m1se; m2s=m2se; m3s=m3se; m4s=m4se; m5s=m5se; m6s=m6se; m7s=m8se;
  prec=prece*10;
  Double err;
  ASYMP_LIMIT_T5a = (abs(ps)+m1s+m2s+m3s+m4s)*1e8;
  T5ex = T5aint(ps,m1s,m2s,m3s,m4s,ASYMP_LIMIT_T5a,prec,&err);
  Cplx z=Cqagpinf2(U7mk,sqr(sqrt(m6s)+sqrt(m7s)),m5s,1,prece*100,&err);
  *abserre = err;
  return(z);
}


/****** Self-energy integrals with double-power propagators: ******/

static Cplx U6m1k(Double x);
static Cplx U6m2k(Double x);
static Cplx U6m3k(Double x);
static Cplx U6n1k(Double x);
static Cplx U6n2k(Double x);


static Cplx U6m2k(Double s1)
{
  Double err1=0;
  Cplx z;
  if(s1 > ASYMP_LIMIT_T5a)
  {
/* Use extrapolation of T5 for large s1, based on form of large-mass expansion
   T5 = (A + B*log(s1) + C*log^2(s1))/s1 + O(s1^-2)
   C = -1/2
   B = -B0fin(ps,m1s,m2s)
   A is determined approximatedly from value of T5 for s1=ASYMP_LIMIT_T5a
*/
    z = T5ex*(ASYMP_LIMIT_T5a/s1) + log(ASYMP_LIMIT_T5a/s1)*
    	(B0A(ps,m1s,m2s)+2 + .5*log(s1) + .5*log(ASYMP_LIMIT_T5a))/s1;
    z *= -DM1DeltaB0(s1,m6s,m7s);
  }
  else
    z = -DM1DeltaB0(s1,m6s,m7s)*(T5aint(ps,m1s,m2s,m3s,m4s,s1,prec,&err1));
  return(z);
}

Cplx U6m2int(Double pse, Double m1se, Double m2se, Double m3se, Double m4se,
             Double m6se, Double m8se, Double prece, Double *abserre)
{
  ps=pse; m1s=m1se; m2s=m2se; m3s=m3se; m4s=m4se; m6s=m6se; m7s=m8se;
  prec=prece*10;
  Double err;
  ASYMP_LIMIT_T5a = (abs(ps)+m1s+m2s+m3s+m4s)*1e8;
  T5ex = T5aint(ps,m1s,m2s,m3s,m4s,ASYMP_LIMIT_T5a,prec,&err);
  Cplx z=Cqaginfasqrtsing2(U6m2k,sqr(sqrt(m6s)+sqrt(m7s)),prece*100,&err);
  *abserre = err;
  return(z);
}


static Cplx U6n1k(Double s1)
{
  Double err1=0;
  Cplx z;
  if(s1 > ASYMP_LIMIT_T5a)
  {
/* Use extrapolation of T5 for large s1, based on form of large-mass expansion
   T5 = (A + B*log(s1) + C*log^2(s1))/s1 + O(s1^-2)
   C = -1
   B = 2-B0fin(ps,m1s,m2s)-B0fin(ps,m4s,m5s)
   A is determined approximatedly from value of T5 for s1=ASYMP_LIMIT_T5a
*/
    z = T5ex*(ASYMP_LIMIT_T5a/s1) + log(ASYMP_LIMIT_T5a/s1)*
    	(B0A(ps,m1s,m2s)+B0A(ps,m4s,m5s)+2 + log(s1) + log(ASYMP_LIMIT_T5a))/s1;
    z *= -DM1DeltaB0(s1,m6s,m7s);
  }
  else
    z = -DM1DeltaB0(s1,m6s,m7s)*(T5aint(ps,m1s,m2s,s1,m4s,m5s,prec,&err1));
  return(z);
}

Cplx U6n1int(Double pse, Double m1se, Double m2se, Double m3se, Double m6se,
             Double m7se, Double m8se, Double prece, Double *abserre)
{
  ps=pse; m1s=m1se; m2s=m2se; m4s=m7se; m5s=m8se; m6s=m3se; m7s=m6se;
  prec=prece*10;
  Double err;
  ASYMP_LIMIT_T5a = (abs(ps)+m1s+m2s+m4s+m5s)*1e8;
  T5ex = T5aint(ps,m1s,m2s,ASYMP_LIMIT_T5a,m4s,m5s,prec,&err);
  Cplx z=Cqaginfasqrtsing2(U6n1k,sqr(sqrt(m6s)+sqrt(m7s)),prece*100,&err);
  *abserre = err;
  return(z);
}
